# frozen_string_literal: true

require_relative 'db/wp_item'
require_relative 'db/updater'
require_relative 'db/wp_items'
require_relative 'db/plugins'
require_relative 'db/themes'
require_relative 'db/plugin'
require_relative 'db/theme'
require_relative 'db/sponsor'
require_relative 'db/wp_version'
require_relative 'db/fingerprints'

require_relative 'db/vuln_api'

require_relative 'db/dynamic_finders/base'
require_relative 'db/dynamic_finders/plugin'
require_relative 'db/dynamic_finders/theme'
require_relative 'db/dynamic_finders/wordpress'
